@@ -161,7 +161,7 @@ module ActiveRecord
   # == <tt>before_validation*</tt> returning statements
   #
   # If the returning value of a +before_validation+ callback can be evaluated to +false+, the process will be aborted and <tt>Base#save</tt> will return +false+.
-  # If <tt>Base#save!</tt> is called it will raise a +RecordNotSaved+ exception.
+  # If Base#save! is called it will raise a RecordNotSaved exception.
   # Nothing will be appended to the errors object.
   #
   # == Canceling callbacks
@@ -183,14 +183,8 @@ module ActiveRecord
         base.send :alias_method_chain, method, :callbacks
       end
 
-      CALLBACKS.each do |method|
-        base.class_eval <<-"end_eval"
-          def self.#{method}(*callbacks, &block)
-            callbacks << block if block_given?
-            write_inheritable_array(#{method.to_sym.inspect}, callbacks)
-          end
-        end_eval
-      end
+      base.send :include, ActiveSupport::Callbacks
+      base.define_callbacks *CALLBACKS
     end
 
     # Is called when the object was instantiated by one of the finders, like <tt>Base.find</tt>.
@@ -235,9 +229,9 @@ module ActiveRecord
     # Is called _after_ <tt>Base.save</tt> on existing objects that have a record.
     def after_update() end
 
-    def update_with_callbacks #:nodoc:
+    def update_with_callbacks(*args) #:nodoc:
       return false if callback(:before_update) == false
-      result = update_without_callbacks
+      result = update_without_callbacks(*args)
       callback(:after_update)
       result
     end
@@ -301,38 +295,15 @@ module ActiveRecord
       def callback(method)
         notify(method)
 
-        callbacks_for(method).each do |callback|
-          result = case callback
-            when Symbol
-              self.send(callback)
-            when String
-              eval(callback, binding)
-            when Proc, Method
-              callback.call(self)
-            else
-              if callback.respond_to?(method)
-                callback.send(method, self)
-              else
-                raise ActiveRecordError, "Callbacks must be a symbol denoting the method to call, a string to be evaluated, a block to be invoked, or an object responding to the callback method."
-              end
-          end
-          return false if result == false
+        result = run_callbacks(method) { |result, object| result == false }
+
+        if result != false && respond_to_without_attributes?(method)
+          result = send(method)
         end
 
-        result = send(method) if respond_to_without_attributes?(method)
-
         return result
       end
 
-      def callbacks_for(method)
-        self.class.read_inheritable_attribute(method.to_sym) or []
-      end
-
-      def invoke_and_notify(method)
-        notify(method)
-        send(method) if respond_to_without_attributes?(method)
-      end
-
       def notify(method) #:nodoc:
         self.class.changed
         self.class.notify_observers(method, self)
